function [ped Fstat Frep] = sfm_predict(ped, groups, walls, T, deltat, ...
					pars, selected_idxs, F)

  N = length(ped);
  X = reshape([ped.x], 2, N);

  if ~exist('selected_idxs', 'var')
    selected_idxs = 1:N;
  end
  
  if ~isempty(walls)
    wall_len = sqrt((walls(:, 3) - walls(:, 1)).^2 + ...
		    (walls(:, 4) - walls(:, 2)).^2); % wall lengths
  else
    wall_len = [];
  end

  t = 0;
  step = 1;
  Fstat = [];
  while t < T

    Frep = zeros(N, N, 2);
    
    for I=selected_idxs
      % === acceleration term ===
      ei0 = ped(I).e0;
      Faccel(I, :) = (ped(I).v0 * ei0 - ped(I).v) / pars.tau;
      
      % === groups term ===
      if groups(I) == 0
	group_idxs = [];
      else
	group_idxs = find(groups == groups(I));
	group_idxs = setdiff(group_idxs, I); % to remove oneself
      end
      Ng = length(group_idxs) + 1;
      thetaA = (Ng-1) / 2;

      if Ng == 1
	Fgroup(I, :) = [0 0];
      else
	% centre of mass of other members
	Ci = mean(X(:, group_idxs), 2);

	% --- fvis ---
	if norm(X(:, I) - Ci) > 0
	  U = (Ci - X(:, I)) / norm(Ci - X(:, I));
	  alpha = vecangle(ped(I).v, U);
	  alpha = alpha - pars.phi;
	  if alpha < 0, alpha = 0; end
	  if isnan(alpha), alpha = 0; end
	else
	  alpha = 0; %pi - pars.phi;
	end
    
	fvis = -pars.beta1 * alpha * ped(I).v;

	% --- fatt ---
	% group center of mass
	C = mean([X(:, group_idxs), X(:, I)], 2); 
	if norm(X(:, I) - C) - ped(I).d0/2 > thetaA
	  fatt = pars.beta2 * U';
	else
	  fatt = [0 0];
	end

	% --- frep ---
	frep = [0 0];
	for J=1:length(group_idxs)
	  dij = norm(X(:, group_idxs(J)) - X(:, I)) - ...
		ped(group_idxs(J)).d0/2 - ped(I).d0/2;
	  if dij < pars.safety_dist
	    W = (X(:, I) - X(:, group_idxs(J)))';
	    W = W / norm(W);
	    frep = frep + pars.beta3 * W;
	  end
	end

	Fgroup(I, :) = fvis + fatt + frep;

	Fstat(I, 1:4, step) = [norm(fvis) norm(fatt) norm(frep) alpha];
      end

      % === wall repulsive force ===
      dw = Inf;
      [x0 y0] = deal(X(1, I), X(2, I));
      for K=1:size(walls, 1)
	[x1 y1 x2 y2] = deal(walls(K, 1), walls(K, 2), ...
			     walls(K, 3), walls(K, 4));
	% distance to the wall 'line'
	dK = abs((x2 - x1) * (y1 - y0) - (x1 - x0) * (y2 - y1)) / ...
	     sqrt((x2 - x1)^2 + (y2 - y1)^2);
	ew = [y2 - y1, -(x2 - x1)];
	ew = ew / norm(ew);
	% distance to the wall 'line segment'
	d1 = norm([x0 y0] - [x1 y1]);
	d2 = norm([x0 y0] - [x2 y2]);
	% ... check whether perpendicular dist is to the wall segment or to
	% the line defined by the wall outside the segment and switch to a
	% point to point distance instead
	if (sqrt(d1^2 - dK^2) > wall_len(K)) || ...
	      sqrt(d2^2 - dK^2) > wall_len(K)
	  if d1 < d2
	    dK = d1;
	    ew = [x0 - x1, y0 - y1];
	  else
	    dK = d2;
	    ew = [x0 - x2, y0 - y2];
	  end
	  ew = ew / norm(ew);
	end
	dK = dK - ped(I).d0/2;
	
	if dK < dw			% save the distance to the nearest wall
	  dw = dK; 
	  ew_best = ew;
	end
      end

      if dw == Inf
	Fwall(I, :) = [0 0];
      else
	Fwall(I, :) = pars.a * exp(-dw/pars.b) * ew_best;
      end

      % === repulsive force between pedestrians ===
      Frepulsive(I, :) = [0 0];
      for J=1:N
	if I == J, continue; end
	
	if groups(I) ~= 0 && groups(I) == groups(J)
	  continue; 
	end
	
	dij = norm(X(:, J) - X(:, I)) - ped(I).d0/2 - ped(J).d0/2;
	% HACK: In real world, two people can get closer than to the sum
        % of their radii. However, the following formula for the
        % repulsive force works only until touching and then produces
        % very quickly very high numbers. This lead to too much strong
        % (unrealistic) force.
	if dij < 0
	  dij = 0;
	end
	eij = (X(:, J) - X(:, I))' / norm(X(:, J) - X(:, I));
	D = pars.lambda * (ped(I).v - ped(J).v) + eij;
	B = pars.gamma * norm(D);
	tij = D / norm(D);
	
	theta = vecangle(tij, eij);
	nij = [-tij(2), tij(1)];
	if theta < pi/2
	  nij = -nij;
	end
	
	fij = -pars.A * exp(-dij/B) * (exp(-(pars.np * B * theta)^2) * tij + ...
				       exp(-(pars.n * B * theta)^2) * nij);
	
	Frepulsive(I, :) = Frepulsive(I, :) + fij;

	Frep(I, J, :) = fij;
	Frep(J, I, :) = -fij;
      end

      % === the final force ===
      F(I, :) = Faccel(I, :) + Fgroup(I, :) + Fwall(I, :) + Frepulsive(I, :);

      Fstat(I, 5:12, step) = [Faccel(I, :) Fgroup(I, :) ...
		    Fwall(I, :) Frepulsive(I, :)];
    end  % for I=1:N

    % velocities and positions update
    for I=1:N
      ped(I).v = ped(I).v + deltat * F(I, :);
      ped(I).x = ped(I).x + deltat * ped(I).v;
    end
    
    t = t + deltat;
    step = step + 1;
  end  % while t < T
  